set client_encoding = 'UTF8';

-- Подключим модуль криптографии для хеширования паролей.
create extension pgcrypto;

-- tabAttribItem

create table tabAttribItem
(
	ItemID bigserial primary key,
	OwnerID bigint null,
	Category uuid null,
	Code varchar(256) null,
	CharValue varchar(256) null,
	IntegerValue integer null,
	FloatValue double precision null,
	DateTimeValue timestamp null,
	PolygonValue polygon null
);

alter table tabAttribItem add constraint FK_tabAttribItem_OwnerID foreign key (OwnerID) references tabAttribItem(ItemID) on update cascade on delete cascade;

create index on tabAttribItem(OwnerID);
create index on tabAttribItem(Category);
create index on tabAttribItem(Code);

create index on tabAttribItem(CharValue);

create index on tabAttribItem(IntegerValue);
create index on tabAttribItem(FloatValue);
create index on tabAttribItem(DateTimeValue);
create index on tabAttribItem using gist(PolygonValue);

-- tabBrowserItem

create table tabBrowserItem
(
	ItemID uuid primary key,
	OwnerID uuid null,
	Category uuid null,
	Name varchar(256) not null,
	CreatedBy uuid not null,
	CreatedAt timestamp not null,
	UpdatedBy uuid not null,
	UpdatedAt timestamp not null,
	Flags smallint null,
	Description text not null,
	LockedBy uuid null,
	InheritsPermissions boolean not null,
	DeletedName text null,
	TotalBlobsSize bigint null
);

alter table tabBrowserItem add constraint FK_tabBrowserItem_OwnerID foreign key (OwnerID) references tabBrowserItem(ItemID) on update cascade on delete cascade;

create index IDX_tabBrowserItem_OwnerID on tabBrowserItem(OwnerID);
create index IDX_tabBrowserItem_Category on tabBrowserItem(Category);
create index IDX_tabBrowserItem_Name on tabBrowserItem(Name);
create index IDX_tabBrowserItem_Name2 on tabBrowserItem(lower(Name) varchar_pattern_ops);

-- tabBrowserItemData

create table tabBrowserItemData
(
	ID bigserial primary key,
	OwnerBrowserItemID uuid not null references tabBrowserItem(ItemID) on update cascade on delete cascade,
	RootAttribItemID bigint not null references tabAttribItem(ItemID),
	IsRevisionData boolean not null,
	CreatedBy uuid null,
	CreatedAt timestamp null,
	RevisionComment text null,
	RevisionIndex smallint null,
	SearchResultCustomInfo text null
);

create index IDX_tabBrowserItemData_OwnerBrowserItemID on tabBrowserItemData(OwnerBrowserItemID);
create index IDX_tabBrowserItemData_RootAttribItemID on tabBrowserItemData(RootAttribItemID);
create index IDX_tabBrowserItemData_RevisionIndex on tabBrowserItemData(RevisionIndex);

-- tabBlob

create table tabBlob
(
	BlobID bigserial primary key,
	Category uuid not null,
	OwnerRevisionID bigint not null references tabBrowserItemData(ID) on update cascade on delete cascade
);

create index IDX_tabBlob_OwnerRevisionID on tabBlob(OwnerRevisionID);

-- tabUnfixedBlobs

create table tabUnfixedBlobs
(
	RecordID bigserial primary key,
	BlobID bigint not null
);

create index IDX_tabUnfixedBlobs_BlobID on tabUnfixedBlobs(BlobID);

-- tabUser

-- Имя пользователя неуникально, т.к. при удалении пользоваель не убирается из БД, а позже может быть добавлен пользователь с таким же именем.
create table tabUser
(
	UserID uuid primary key,
	Name varchar(256) not null, 
	Login varchar(256) unique null,
	Password varchar(256) null,
	WindowsSID varchar(256) unique null,
	Locked boolean,
	Deleted boolean,
	ContactInfo text
);

alter table tabBrowserItem add constraint FK_tabBrowserItem_CreatedBy foreign key (CreatedBy) references tabUser (UserID) on update cascade on delete restrict;
alter table tabBrowserItem add constraint FK_tabBrowserItem_UpdatedBy foreign key (UpdatedBy) references tabUser (UserID) on update cascade on delete restrict;
alter table tabBrowserItem add constraint FK_tabBrowserItem_LockedBy foreign key (LockedBy) references tabUser (UserID) on update cascade on delete restrict;

alter table tabBrowserItemData add constraint FK_tabBrowserItemData_CreatedBy foreign key (CreatedBy) references tabUser (UserID) on update cascade on delete restrict;

-- tabGroup

create table tabGroup
(
	GroupID uuid primary key,
	Name varchar(256) unique not null,
	Description text not null,
	IsPublic boolean
);

-- tabUserGroup

create table tabUserGroup
(
	UserID uuid not null,
	GroupID uuid not null
);

alter table tabUserGroup add constraint PK_tabUserGroup primary key (UserID, GroupID);
alter table tabUserGroup add constraint FK_tabUserGroup_UserID foreign key (UserID) references tabUser (UserID) on update cascade on delete cascade;
alter table tabUserGroup add constraint FK_tabUserGroup_GroupID foreign key (GroupID) references tabGroup (GroupID) on update cascade on delete cascade;

-- tabAccessControl

create table tabAccessControl
(
	BrowserItemID uuid not null,
	GroupID uuid not null,
	Permissions integer not null
);

alter table tabAccessControl add constraint PK_tabAccessControl primary key (BrowserItemID, GroupID);
alter table tabAccessControl add constraint FK_tabAccessControl_BrowserItemID foreign key (BrowserItemID) references tabBrowserItem (ItemID) on update cascade on delete cascade;
alter table tabAccessControl add constraint FK_tabAccessControl_GroupID foreign key (GroupID) references tabGroup (GroupID) on update cascade on delete cascade;

-- tabOperationsLog

create table tabOperationsLog
(
	ID bigserial primary key,
	OpTimeStamp timestamp not null,
	UserID uuid not null references tabUser(UserID) on update cascade on delete restrict,
	Message text not null
);

create index on tabOperationsLog(OpTimeStamp);

-- tabVersion

create table tabVersion
(
	Version integer not null
);

-- Создаем встроенные объекты безопасности.

-- UUID групп используется в исходных кодах серверной части! Не менять!
insert into tabGroup (GroupID, Name, Description)
values ('dbfb2161-4b0d-480e-92d6-b6b3dc82ebf6', 'Администраторы', 'Встроенная группа администраторов.');

-- UUID администратора, его логин и пароль используются
-- в исходных кодах серверной части! Не менять!
insert into tabUser (UserID, Name, Login, Password, Locked, Deleted)
values ('3c7cb0ce-e47f-4dd1-94f2-d20217b15c1e', 'Администратор', 'root', crypt('', gen_salt('md5')), false, false);

-- Добавим администратора в группы Administrators.
insert into tabUserGroup (UserID, GroupID)
values ('3c7cb0ce-e47f-4dd1-94f2-d20217b15c1e', 'dbfb2161-4b0d-480e-92d6-b6b3dc82ebf6');

-- Идентификатор корневой папки: 8875bb41-4413-4c59-8c3f-cdb90c65fc77
-- Используются в исходных кодах серверной части! Не менять!

-- Создадим корневую папку.
insert into tabBrowserItem(ItemID, OwnerID, Category, Name, CreatedBy, CreatedAt, UpdatedBy, UpdatedAt, InheritsPermissions, Description)
values (
	'8875bb41-4413-4c59-8c3f-cdb90c65fc77', null, null, '',
	'3c7cb0ce-e47f-4dd1-94f2-d20217b15c1e', now() at time zone 'utc',
	'3c7cb0ce-e47f-4dd1-94f2-d20217b15c1e', now() at time zone 'utc', false, '');

-- Дадим полный доступ security_permissions::SpFullAccess.
-- Флаги доступа смотри в src/server/storage/security_permissions.h.
insert into tabAccessControl(BrowserItemID, GroupID, Permissions)
values ('8875bb41-4413-4c59-8c3f-cdb90c65fc77', 'dbfb2161-4b0d-480e-92d6-b6b3dc82ebf6', 7);

-- Установим версию.
insert into tabVersion(Version)
values (3);

-- Триггеры.

-- При удалении tabBlob'a надо его отметить для garbage collector'a.
create function DoUnfixBlob() returns trigger as $body$
begin
	insert into tabUnfixedBlobs(BlobID) values(old.BlobID);
	return null; -- result is ignored since this is an AFTER trigger
end;
$body$ language plpgsql;

create trigger UnfixBlob after delete on tabBlob
	for each row execute procedure DoUnfixBlob();

-- Удаление дерева attrib item'ов, при удаление tabBrowserItemData.
create function DoRemoveAttribItemTree() returns trigger as $body$
begin
  delete from tabAttribItem tai where tai.ItemID = old.RootAttribItemID;
  return null; -- result is ignored since this is an AFTER trigger
end;
$body$ language plpgsql;

create trigger RemoveAttribItemTree after delete on tabBrowserItemData
  for each row execute procedure DoRemoveAttribItemTree();

-- Построение пути вида 'путь/из/uuid:/путь/для/пользователя"
-- Пример результата:
-- "8875bb41-4413-4c59-8c3f-cdb90c65fc77/a8b36343-121c-43d2-87e9-e6e75cc39e04/727bd50f-7fcf-4785-adc0-b945408f22de/2af73b92-7751-40a5-a70b-5c01e200dc25:/Folder_0_3/Folder_01_2/File_012_2"
create function GetBrowserItemPath(bwid uuid) returns text as $$
declare
	result text;
begin
	with recursive Hierarchy(ItemID, OwnerID, Name, Level, Path, NamePath) as
	(
		select tbi.ItemID, tbi.OwnerID, tbi.Name, 1 as Level, cast(tbi.ItemID as text) as Path, '/' || cast(tbi.Name as text) as NamePath
		from tabBrowserItem as tbi
		where tbi.ItemID = bwid
		union all
		select tbi.ItemID, tbi.OwnerID, tbi.Name, eh.Level + 1 as Level,
			tbi.ItemID || '/' || eh.Path as Path,
			case when tbi.ItemID <> '8875bb41-4413-4c59-8c3f-cdb90c65fc77' then '/' else '' end || tbi.Name || eh.NamePath as NamePath
		from tabBrowserItem as tbi
		inner join Hierarchy eh on eh.OwnerID = tbi.ItemID
	)
	select Path || ':' || NamePath into result from Hierarchy order by Level desc limit 1;
	return result;
end;
$$ language plpgsql;

create function UserHasPermissionsOnBrowserItem(bwid uuid, uid uuid, perm integer) returns boolean as $$
declare
	result boolean;
begin
	select bit_or(Permissions) is not null and bit_or(Permissions) & perm = perm into result
	from tabAccessControl where BrowserItemID = bwid and GroupID in (select GroupID from tabUserGroup where UserID = uid);
	return result;
end;
$$ language plpgsql;